@@ -112,6 +112,8 @@ bool ibx_partial_backup = false;
 char *ibx_position_arg = NULL;
 char *ibx_backup_directory = NULL;
 
+extern bool xb_opt_destroy_password;
+
 /* copy of proxied xtrabackup options */
 my_bool ibx_xb_close_files;
 const char *ibx_xtrabackup_compress_alg;
@@ -214,7 +216,7 @@ ibx_mode_t ibx_mode = IBX_MODE_BACKUP;
 
 static struct my_option ibx_long_options[] =
 {
-	{"version", 'v', "print version information",
+	{"version", 'v', "print xtrabackup version information",
 	 (uchar *) &opt_ibx_version, (uchar *) &opt_ibx_version, 0,
 	 GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
 
@@ -257,11 +259,12 @@ static struct my_option ibx_long_options[] =
 	 (uchar *) &opt_ibx_slave_info, (uchar *) &opt_ibx_slave_info, 0,
 	 GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
 
-	{"incremental", OPT_INCREMENTAL,
-	 "Create an incremental backup, rather than a full one. When this option is specified, "
+	{"incremental", OPT_INCREMENTAL, "This option tells xtrabackup to "
+	 "create an incremental backup, rather than a full one. It is passed "
+	 "to the xtrabackup child process. When this option is specified, "
 	 "either --incremental-lsn or --incremental-basedir can also be given. "
-	 "If neither option is given, option --incremental-basedir is used "
-	 "by default, set to the first timestamped backup "
+	 "If neither option is given, option --incremental-basedir is passed "
+	 "to xtrabackup by default, set to the first timestamped backup "
 	 "directory in the backup base directory.",
 	 (uchar *) &opt_ibx_incremental, (uchar *) &opt_ibx_incremental, 0,
 	 GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
@@ -376,14 +379,14 @@ static struct my_option ibx_long_options[] =
 	{"incremental-history-name", OPT_INCREMENTAL_HISTORY_NAME,
 	 "This option specifies the name of the backup series stored in the "
 	 "PERCONA_SCHEMA.xtrabackup_history history record to base an "
-	 "incremental backup on. Backup will search the history table "
+	 "incremental backup on. Xtrabackup will search the history table "
 	 "looking for the most recent (highest innodb_to_lsn), successful "
 	 "backup in the series and take the to_lsn value to use as the "
 	 "starting lsn for the incremental backup. This will be mutually "
 	 "exclusive with --incremental-history-uuid, --incremental-basedir "
 	 "and --incremental-lsn. If no valid lsn can be found (no series by "
-	 "that name, no successful backups by that name), "
-	 "an error will be returned. It is used with the --incremental option.",
+	 "that name, no successful backups by that name) xtrabackup will "
+	 "return with an error. It is used with the --incremental option.",
 	 (uchar*) &opt_ibx_incremental_history_name,
 	 (uchar*) &opt_ibx_incremental_history_name, 0, GET_STR,
 	 REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
@@ -393,8 +396,8 @@ static struct my_option ibx_long_options[] =
 	 "stored in the PERCONA_SCHEMA.xtrabackup_history to base an "
 	 "incremental backup on. --incremental-history-name, "
 	 "--incremental-basedir and --incremental-lsn. If no valid lsn can be "
-	 "found (no success record with that uuid), an error will be returned."
-	 " It is used with the --incremental option.",
+	 "found (no success record with that uuid) xtrabackup will return "
+	 "with an error. It is used with the --incremental option.",
 	 (uchar*) &opt_ibx_incremental_history_uuid,
 	 (uchar*) &opt_ibx_incremental_history_uuid, 0, GET_STR,
 	 REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
@@ -423,7 +426,7 @@ static struct my_option ibx_long_options[] =
 	{"include", OPT_INCLUDE,
 	 "This option is a regular expression to be matched against table "
 	 "names in databasename.tablename format. It is passed directly to "
-	 "--tables option. See the documentation for "
+	 "xtrabackup's --tables option. See the xtrabackup documentation for "
 	 "details.",
 	 (uchar*) &opt_ibx_include,
 	 (uchar*) &opt_ibx_include, 0, GET_STR,
@@ -473,6 +476,12 @@ static struct my_option ibx_long_options[] =
 	 (uchar*) &opt_ibx_lock_wait_threshold, 0, GET_UINT,
 	 REQUIRED_ARG, 60, 0, 0, 0, 0, 0},
 
+	{"debug-sleep-before-unlock", OPT_DEBUG_SLEEP_BEFORE_UNLOCK,
+	 "This is a debug-only option used by the XtraBackup test suite.",
+	 (uchar*) &opt_ibx_debug_sleep_before_unlock,
+	 (uchar*) &opt_ibx_debug_sleep_before_unlock, 0, GET_UINT,
+	 REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
+
 	{"safe-slave-backup-timeout", OPT_SAFE_SLAVE_BACKUP_TIMEOUT,
 	 "How many seconds --safe-slave-backup should wait for "
 	 "Slave_open_temp_tables to become zero. (default 300)",
@@ -485,20 +494,22 @@ static struct my_option ibx_long_options[] =
 	We put them here with only purpose for them to showup in
 	innobackupex --help output */
 
-	{"close_files", OPT_CLOSE_FILES, "Do not keep files opened."
-	" Use at your own risk.",
+	{"close_files", OPT_CLOSE_FILES, "Do not keep files opened. This "
+	 "option is passed directly to xtrabackup. Use at your own risk.",
 	 (uchar*) &ibx_xb_close_files, (uchar*) &ibx_xb_close_files, 0,
 	 GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
 
-	{"compress", OPT_COMPRESS, "This option instructs backup to "
-	 "compress backup copies of InnoDB data files."
-	 , (uchar*) &ibx_xtrabackup_compress_alg,
+	{"compress", OPT_COMPRESS, "This option instructs xtrabackup to "
+	 "compress backup copies of InnoDB data files. It is passed directly "
+	 "to the xtrabackup child process. Try 'xtrabackup --help' for more "
+	 "details.", (uchar*) &ibx_xtrabackup_compress_alg,
 	 (uchar*) &ibx_xtrabackup_compress_alg, 0,
 	 GET_STR, OPT_ARG, 0, 0, 0, 0, 0, 0},
 
 	{"compress-threads", OPT_COMPRESS_THREADS,
 	 "This option specifies the number of worker threads that will be used "
-	 "for parallel compression.",
+	 "for parallel compression. It is passed directly to the xtrabackup "
+	 "child process. Try 'xtrabackup --help' for more details.",
 	 (uchar*) &ibx_xtrabackup_compress_threads,
 	 (uchar*) &ibx_xtrabackup_compress_threads,
 	 0, GET_UINT, REQUIRED_ARG, 1, 1, UINT_MAX, 0, 0, 0},
@@ -509,15 +520,17 @@ static struct my_option ibx_long_options[] =
 	 (uchar*) &ibx_xtrabackup_compress_chunk_size,
 	 0, GET_ULL, REQUIRED_ARG, (1 << 16), 1024, ULONGLONG_MAX, 0, 0, 0},
 
-	{"export", OPT_EXPORT, " enables exporting individual tables for import "
-	 "into another server.",
+	{"export", OPT_EXPORT, "This option is passed directly to xtrabackup's "
+	 "--export option. It enables exporting individual tables for import "
+	 "into another server. See the xtrabackup documentation for details.",
 	 (uchar*) &ibx_xtrabackup_export, (uchar*) &ibx_xtrabackup_export,
 	 0, GET_BOOL, NO_ARG, 0, 0, 0, 0, 0, 0},
 
 	{"extra-lsndir", OPT_EXTRA_LSNDIR, "This option specifies the "
 	 "directory in which to save an extra copy of the "
 	 "\"xtrabackup_checkpoints\" file. The option accepts a string "
-	 "argument.",
+	 "argument. It is passed directly to xtrabackup's --extra-lsndir "
+	 "option. See the xtrabackup documentation for details.",
 	 (uchar*) &ibx_xtrabackup_extra_lsndir,
 	 (uchar*) &ibx_xtrabackup_extra_lsndir,
 	 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
@@ -539,7 +552,7 @@ static struct my_option ibx_long_options[] =
 	 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
 
 	{"incremental-force-scan", OPT_INCREMENTAL_FORCE_SCAN,
-	 "Perform full scan of data files "
+	 "This options tells xtrabackup to perform full scan of data files "
 	 "for taking an incremental backup even if full changed page bitmap "
 	 "data is available to enable the backup without the full scan.",
 	 (uchar*)&ibx_xtrabackup_incremental_force_scan,
@@ -565,8 +578,10 @@ static struct my_option ibx_long_options[] =
 	 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
 
 	{"parallel", OPT_PARALLEL, "On backup, this option specifies the "
-	 "number of threads to use to back "
-	 "up files concurrently.  The option accepts an integer argument.",
+	 "number of threads the xtrabackup child process should use to back "
+	 "up files concurrently.  The option accepts an integer argument. It "
+	 "is passed directly to xtrabackup's --parallel option. See the "
+	 "xtrabackup documentation for details.",
 	 (uchar*) &ibx_xtrabackup_parallel, (uchar*) &ibx_xtrabackup_parallel,
 	 0, GET_INT, REQUIRED_ARG, 1, 1, INT_MAX, 0, 0, 0},
 
@@ -574,21 +589,23 @@ static struct my_option ibx_long_options[] =
 	{"stream", OPT_STREAM, "This option specifies the format in which to "
 	 "do the streamed backup.  The option accepts a string argument. The "
 	 "backup will be done to STDOUT in the specified format. Currently, "
-	 "the only supported formats are tar and mbstream/xbstream.",
+	 "the only supported formats are tar and mbstream/xbstream. This "
+	 "option is passed directly to xtrabackup's --stream option.",
 	 (uchar*) &ibx_xtrabackup_stream_str,
 	 (uchar*) &ibx_xtrabackup_stream_str, 0, GET_STR,
 	 REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
 
 	{"tables-file", OPT_TABLES_FILE, "This option specifies the file in "
 	 "which there are a list of names of the form database.  The option "
-	 "accepts a string argument.table, one per line.",
+	 "accepts a string argument.table, one per line. The option is passed "
+	 "directly to xtrabackup's --tables-file option.",
 	 (uchar*) &ibx_xtrabackup_tables_file,
 	 (uchar*) &ibx_xtrabackup_tables_file,
 	 0, GET_STR, REQUIRED_ARG, 0, 0, 0, 0, 0, 0},
 
 	{"throttle", OPT_THROTTLE, "This option specifies a number of I/O "
 	 "operations (pairs of read+write) per second.  It accepts an integer "
-	 "argument.",
+	 "argument.  It is passed directly to xtrabackup's --throttle option.",
 	 (uchar*) &ibx_xtrabackup_throttle, (uchar*) &ibx_xtrabackup_throttle,
 	 0, GET_LONG, REQUIRED_ARG, 0, 0, LONG_MAX, 0, 1, 0},
 
@@ -600,10 +617,11 @@ static struct my_option ibx_long_options[] =
 	 0, 0, 0, 0, 0, 0},
 
 	{"use-memory", OPT_USE_MEMORY, "This option accepts a string argument "
-	 "that specifies the amount of memory in bytes to use "
+	 "that specifies the amount of memory in bytes for xtrabackup to use "
 	 "for crash recovery while preparing a backup. Multiples are supported "
 	 "providing the unit (e.g. 1MB, 1GB). It is used only with the option "
-	 "--apply-log.",
+	 "--apply-log. It is passed directly to xtrabackup's --use-memory "
+	 "option. See the xtrabackup documentation for details.",
 	 (uchar*) &ibx_xtrabackup_use_memory,
 	 (uchar*) &ibx_xtrabackup_use_memory,
 	 0, GET_LL, REQUIRED_ARG, 100*1024*1024L, 1024*1024L, LONGLONG_MAX, 0,
@@ -625,7 +643,7 @@ static struct my_option ibx_long_options[] =
 
 static void usage(void)
 {
-	puts("Open source backup tool\n\
+	puts("Open source backup tool for InnoDB and XtraDB\n\
 \n\
 Copyright (C) 2009-2015 Percona LLC and/or its affiliates.\n\
 Portions Copyright (C) 2000, 2011, MySQL AB & Innobase Oy. All Rights Reserved.\n\
@@ -676,7 +694,7 @@ innobackupex [--decompress]\n\
 \n\
 DESCRIPTION\n\
 \n\
-The first command line above makes a hot backup of a database.\n\
+The first command line above makes a hot backup of a MySQL database.\n\
 By default it creates a backup directory (named by the current date\n\
 	and time) in the given backup root directory.  With the --no-timestamp\n\
 option it does not create a time-stamped backup directory, but it puts\n\
@@ -686,18 +704,22 @@ indexes in all databases or in all of the databases specified with the\n\
 --databases option.  The created backup contains .frm, .MRG, .MYD,\n\
 .MYI, .MAD, .MAI, .TRG, .TRN, .ARM, .ARZ, .CSM, CSV, .opt, .par, and\n\
 InnoDB data and log files.  The MY.CNF options file defines the\n\
-location of the database.\n\
+location of the database.  This command connects to the MySQL server\n\
+using the mysql client program, and runs xtrabackup as a child\n\
+process.\n\
 \n\
 The --apply-log command prepares a backup for starting a MySQL\n\
 server on the backup. This command recovers InnoDB data files as specified\n\
 in BACKUP-DIR/backup-my.cnf using BACKUP-DIR/ib_logfile0,\n\
 and creates new InnoDB log files as specified in BACKUP-DIR/backup-my.cnf.\n\
-The BACKUP-DIR should be the path to a backup directory\n\
+The BACKUP-DIR should be the path to a backup directory created by\n\
+xtrabackup. This command runs xtrabackup as a child process, but it does not \n\
+connect to the database server.\n\
 \n\
 The --copy-back command copies data, index, and log files\n\
 from the backup directory back to their original locations.\n\
 The MY.CNF options file defines the original location of the database.\n\
-The BACKUP-DIR is the path to a backup directory.\n\
+The BACKUP-DIR is the path to a backup directory created by xtrabackup.\n\
 \n\
 The --move-back command is similar to --copy-back with the only difference that\n\
 it moves files to their original locations rather than copies them. As this\n\
@@ -763,17 +785,7 @@ ibx_get_one_option(int optid,
 		xtrabackup_compress = TRUE;
 		break;
 	case 'p':
-		if (argument)
-		{
-		        char *start = argument;
-			my_free(opt_ibx_password);
-			opt_ibx_password= my_strdup(argument, MYF(MY_FAE));
-			/*  Destroy argument */
-			while (*argument)
-				*argument++= 'x';
-			if (*start)
-				start[1]=0 ;
-		}
+    opt_ibx_password= argument;
 		break;
         }
 	return(0);
